<?php

declare(strict_types=1);

namespace apexl\Io\interfaces;

interface RelationshipViaPivotInterface
{
    public function getLocalTable(): string;

    public function getLocalKey(): string;

    public function getPivotTable(): string;

    public function getPivotTableAlias(): string;

    public function getPivotLocal(): string;

    public function getPivotFields(): array;
}