<?php

declare(strict_types=1);

namespace apexl\Io\interfaces;

use apexl\Io\factories\EntityFactory;

interface HasEntityFactory
{
    public static function factory(): EntityFactory;

    /**
     * @return class-string<EntityFactory>
     */
    public static function entityFactoryBaseClass(): string;
}