<?php

namespace apexl\Io\interfaces;

use Deprecated;

interface HasServicesInterface
{
    #[Deprecated (message: 'Use `services.php` definition file in module root')]
    public function addServices(): array;
}
